﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Resources;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Algorithms.Bandits;
using HeuristicLab.Algorithms.Bandits.GrammarPolicies;
using HeuristicLab.Algorithms.MonteCarloTreeSearch;
using HeuristicLab.Algorithms.MonteCarloTreeSearch.Expansion;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.GrammaticalOptimization.Solvers
{
    public class MonteCarloTreeSearch : SolverBase
    {
        private readonly int maxLen;
        private readonly IProblem problem;
        private readonly Random random;
        private readonly IBanditPolicy behaviourPolicy;
        private readonly IExpansionPolicy expansionPolicy;
        private readonly ISimulationPolicy simulationPolicy;
        private TreeNode rootNode;
        private List<IBanditPolicyActionInfo> actions;
        private List<TreeNode> nodes;

        public MonteCarloTreeSearch(IProblem problem, int maxLen, Random random, IBanditPolicy behaviourPolicy,
            IExpansionPolicy expansionPolicy, ISimulationPolicy simulationPolicy)
        {
            this.problem = problem;
            this.maxLen = maxLen;
            this.random = random;
            this.behaviourPolicy = behaviourPolicy;
            this.expansionPolicy = expansionPolicy;
            this.simulationPolicy = simulationPolicy;
        }

        public bool StopRequested
        {
            get;
            set;
        }

        public override void Run(int maxIterations)
        {
            Reset();
            for (int i = 0; !StopRequested && !Done() && i < maxIterations; i++)
            {
                // select by behaviour policy
                TreeNode currentNode;
                do
                {
                    int currentActionIndex = behaviourPolicy.SelectAction(random, actions);
                    currentNode = nodes[currentActionIndex];
                } while (!Expandable(currentNode));

                // expand tree
                currentNode = expansionPolicy.ExpandTreeNode(currentNode);
                // simulate
                double reward = simulationPolicy.Simulate(currentNode);
                // propagate/reward
                Propagate(currentNode, reward);
            }
        }

        private void Reset()
        {
            StopRequested = false;
            bestQuality = 0.0;
            rootNode = new TreeNode(null, problem.Grammar.SentenceSymbol.ToString(), true, new List<int>() { 0 });
        }

        private bool Done()
        {
            return !rootNode.expandable;
        }

        private bool Expandable(TreeNode node)
        {
            return !problem.Grammar.IsTerminal(node.phrase);
        }

        private void Propagate(TreeNode node, double reward)
        {
            var currentNode = node;
            do
            {
                currentNode.actionInfo.UpdateReward(reward);
                currentNode = node.parent;
            } while (currentNode != null);
        }
    }
}
